// Created by inigo quilez - iq/2016 https://www.shadertoy.com/view/Xtt3Wn
// License Creative Commons Attribution-NonCommercial-ShareAlike 3.0
// More info here: http://iquilezles.org/www/articles/simplepathtracing/simplepathtracing.htm

//adapt for Fragmentaium by CJR (^_^)
//all rands in 3D-1.frag
#info define WANGHASH for best quality
//need speculars, soft shadows, id's, material mode...

#include "3D-1.frag" 

#group Raytracer
uniform int RayBounces;slider[1,5,128]
uniform int RaySteps;slider[64,128,4000]
uniform float FudgeFactor;slider[0.1,1.0,1.0]
uniform float HitDistance;slider[0.00001,0.00018,0.001]
//uniform float SoftShadows;slider[0.0,1.0,1.0]
uniform float MaxDepth;slider[1,15,200]

#group Material
uniform vec3 Mat_Color; color[.65,.65,.65]
uniform float Mat_Refl;slider[0.0,0.0,1.0]
uniform float Mat_ReflExp;slider[0.00001,.75,1.0]
uniform float Mat_Spec;slider[0.0,0.0,10.0]
uniform float Mat_SpecExp;slider[.1,1.0,500.0]
uniform bool UseTexture; checkbox[false]
uniform sampler2D DeTex; file[tex.jpg]
uniform float DeTexScale;slider[0.0,0.5,10.0]
uniform float DeTexExposure;slider[0.0,1.0,2.0]

#group Coloring
vec4 orbitTrap = vec4(10000.0);
uniform float OrbitStrength; slider[0,0,1]
uniform vec4 X; color[-1,0.7,1,0.5,0.6,0.6];
uniform vec4 Y; color[-1,0.4,1,1.0,0.6,0.0];
uniform vec4 Z; color[-1,0.5,1,0.8,0.78,1.0];
uniform vec4 R; color[-1,0.12,1,0.4,0.7,1.0];
uniform bool CycleColors; checkbox[false]
uniform float Cycles; slider[0.1,1.1,32.3]

#group Environment
uniform vec4 SpotLight; color[0.0,1.,10.0,1.0,.8,.5];
uniform vec2 SpotLightDir;  slider[(-1,-1),(0.1,0.1),(1,1)]
uniform float SpotSize;slider[0.0,0.01,10.0]
uniform float SpotExp;slider[0.0001,0.5,20.0]
uniform vec4 BgColor; color[0.0,1.,10.0,.3,.5,.7];
uniform float Ambience;slider[0.0,1.0,5.0]
uniform float Fog;slider[0.0,0.0,1.5]

#group GodRays
uniform int GRIter;slider[0,3,20]
uniform float GRPower; slider[0.0,.3,1.0]

vec3 cosineDirection(in vec3 nor)
{//return a random direction on the hemisphere
	vec2 r = rand2n()*TWO_PI;
	vec3 dr=vec3(sin(r.x)*vec2(sin(r.y),cos(r.y)),cos(r.x));
	return (dot(dr,nor)<0.0)?-dr:dr;
}

vec3 uniformVector( in float seed)
{
    float a = PI*rand( 78.233 + seed);
    float b = TWO_PI*rand( 10.873 + seed);
    return vec3( sin(b)*sin(a), cos(b)*sin(a), cos(a) );
}

vec4 texcube( sampler2D sam, in vec3 p, in vec3 n )
{
    vec3 m = n*n;
	vec4 x = texture2D( sam, p.yz );
	vec4 y = texture2D( sam, p.zx );
	vec4 z = texture2D( sam, p.xy );
	return (x*m.x + y*m.y + z*m.z) / (m.x+m.y+m.z);
}

float DE(vec3 pos);

float map(vec3 z){
	return DE(z);
}

vec3 calcNormal( in vec3 pos )
{
    vec3 eps = vec3(0.0001,0.0,0.0);
    return normalize( vec3(
      map( pos+eps.xyy ) - map( pos-eps.xyy ),
      map( pos+eps.yxy ) - map( pos-eps.yxy ),
      map( pos+eps.yyx ) - map( pos-eps.yyx ) ) );
}

float intersect( in vec3 ro, in vec3 rd )
{
    float res = -1.0;
	float tmax = MaxDepth;
    float t = 0.01;
    for(int i=0; i<RaySteps; i++ )
    {
        float h = map(ro+rd*t)*FudgeFactor;
		if( h<HitDistance || t>tmax ) break;
        t +=  h;
    }
    if( t<tmax ) res = t;
    return res;
}

float shadow( in vec3 ro, in vec3 rd )
{
    float res = 0.0;
    float tmax = 17.0;
    float t = 0.001;
	float sa = rand(viewCoord*(float(subframe)+1.0));
    for(int i=0; i<150; i++ )
    {
		float h = map(ro+rd*t)*FudgeFactor;
        if( h<0.0001 || t>tmax) break;
        t += h;
    }
    if( t>tmax ) res = 1.0;
    return res;
}

vec3 cycle(vec3 c, float s) {
	return vec3(0.5)+0.5*vec3(cos(s*Cycles+c.x),cos(s*Cycles+c.y),cos(s*Cycles+c.z));
}
vec3 getColor() {
	orbitTrap.w = sqrt(orbitTrap.w);
	
	vec3 orbitColor;
	if (CycleColors) {
		orbitColor = cycle(X.xyz,orbitTrap.x)*X.w*orbitTrap.x +
		cycle(Y.xyz,orbitTrap.y)*Y.w*orbitTrap.y +
		cycle(Z.xyz,orbitTrap.z)*Z.w*orbitTrap.z +
		cycle(R.xyz,orbitTrap.w)*R.w*orbitTrap.w;
	} else {
		orbitColor = X.xyz*X.w*orbitTrap.x +
		Y.xyz*Y.w*orbitTrap.y +
		Z.xyz*Z.w*orbitTrap.z +
		R.xyz*R.w*orbitTrap.w;
	}
	
	vec3 color = mix(Mat_Color, 3.0*orbitColor,  OrbitStrength);
	return color;
}

vec3 color(vec3 ro, vec3 rd)
{
    const float epsilon = 0.0001;
    vec3 colorMask = vec3(1.0);
    vec3 accumulatedColor = vec3(0.0);
    vec3 oro = ro;
    vec3 ord = rd;
    float fdis = 0.0;
	float sa = rand(viewCoord*(float(subframe)+1.0));
	vec3 sunDir = vec3(sin(SpotLightDir.x*PI)*cos(SpotLightDir.y*PI/2.0), sin(SpotLightDir.y*PI/2.0)*sin(SpotLightDir.x*PI), cos(SpotLightDir.x*PI));
	sunDir = normalize(sunDir);
	vec3 sunCol = SpotLight.xyz*SpotLight.w; 
	vec3 skyCol =  BgColor.xyz*BgColor.w;
	if(SpotLight.w!=0)skyCol+=pow(max(0.0,dot(rd,sunDir))*1.001  ,SpotExp*100)*SpotSize*SpotLight.xyz;
	
    for( int bounce = 0; bounce<RayBounces; bounce++ ) // bounces of GI
    {
        rd = normalize(rd);
        // trace
        float t = intersect( ro, rd );
        if( t < 0.0 ){
            if( bounce==0 ){ 
				accumulatedColor=skyCol;
				//if(SpotLight.w!=0)accumulatedColor+=pow(max(0.0,dot(rd,sunDir))*1.001  ,SpotExp*100)*SpotSize*SpotLight.xyz;
				break;
			}
            break;
        }
        if( bounce==0 ) fdis = t;
        vec3 pos = ro + rd * t;
        vec3 nor = calcNormal( pos );
		vec3 surfaceColor = clamp(getColor(),0.0,1.0);
        if(UseTexture)surfaceColor+=clamp(texcube(DeTex, pos*DeTexScale, nor ).xyz,0.0,1.0)*DeTexExposure;
       
        // add direct lighitng
        colorMask *= surfaceColor;
        vec3 iColor = vec3(0.0);
//---------------------------------------
		// An attempt at Physcical Based Specular Shading:
		// http://renderwonk.com/publications/s2010-shading-course/
		// (Blinn-Phong with Schickl term and physical normalization)
		/*
		float nDotL = max(0.0,dot(n,sunDir));
		vec3 halfVector = normalize(-dir+sunDir);
		float hDotN = max(0.,dot(n,halfVector));
		float f0=(SpecularExp-1.)/(SpecularExp+1.); f0*=f0;
		float fresnel=f0+(1.-f0)*pow(1.+dot(n,dir),5.);
		float specular =((SpecularExp+2.)/8.)*fresnel*nDotL*pow(hDotN,SpecularExp+0.00001)*Specular;			
		*/
		vec3 n = nor;
		vec3 dir = rd;
		float nDotL = max(0.0,dot(n,sunDir));
		vec3 halfVector = normalize(-dir+sunDir);
		float hDotN = max(0.,dot(n,halfVector));
		float f0=(Mat_SpecExp-1.)/(Mat_SpecExp+1.); f0*=f0;
		float fresnel=f0+(1.-f0)*pow(1.+dot(n,dir),5.);
		float specular =((Mat_SpecExp+2.)/8.)*fresnel*nDotL*pow(hDotN,Mat_SpecExp+0.00001)*Mat_Spec;			
//---------------------------------------
		// light 1        
        float sunDif =  max(0.0, dot(sunDir, nor));
        float sunSha = 1.0; if( sunDif > 0.00001 ) sunSha = shadow( pos + nor*epsilon, sunDir);
        iColor += sunCol * sunDif * sunSha+specular*sunSha*sunDif*sunCol;
        // todo - add back direct specular
		
        // light 2
		vec3 skyPoint = cosineDirection(nor);
        float skySha = shadow( pos + nor*epsilon, skyPoint);
		iColor += BgColor.xyz*skySha*Ambience;
		//iColor += skyCol*skySha*Ambience;
        accumulatedColor += colorMask * iColor;
		//accumulatedColor += colorMask*iColor+colorMask*iColor*specular*sunSha;
		
        // calculate new ray
        float isDif = 1.-Mat_Refl;
        if( rand(sa + 1.123 + 7.7*float(bounce)) < isDif ){
			rd = cosineDirection(nor);//rd = cosineDirection(76.2 + 73.1*float(bounce) + sa + 17.7*float(subframe), nor);
        } else {
			rd=normalize(reflect(rd, nor))+uniformVector(sa + 111.123 + 65.2*float(bounce))*(1.-Mat_ReflExp);
        }
        ro = pos;
   }
   float ff = exp(-0.1*fdis);
   accumulatedColor *= ff; 
 
   // volumetrics
   float acc = 0.0;
   vec3 pos = oro + ord;
   for( int i=0; i<GRIter; i++ )
   {
	   float t = fdis*rand(sa+13.731*float(i)+float(subframe)*7.773);
       vec3 pos = oro + ord*t;
       acc += GRPower*shadow( pos, sunDir );//*exp(-0.25*t);
	}
    accumulatedColor += vec3(0.1)*pow(acc,2.0)*sunCol*0.4;
	float fg=1.0-exp(-pow(Fog,4.0)*fdis*fdis); //simple fog
	accumulatedColor = mix(accumulatedColor, skyCol, fg);
return accumulatedColor;
}
